// 67d7842dbbe25473c3c32b93c0da8047785f30d78e8a024de1b57352245f9689
/*******************************************************************************
 * Copyright (c) 2007, 2012 Wind River Systems, Inc. and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License 2.0
 * which accompanies this distribution, and is available at
 * https://www.eclipse.org/legal/epl-2.0/
 *
 * Contributors:
 *     Wind River Systems - initial API and implementation
 *******************************************************************************/

/*******************************************************************************
 * Modified by Xilinx,Inc
 * Copyright 2022 Xilinx, Inc. All Rights Reserved. Modifications
 * are made available under the terms of the Eclipse Public License 2.0
 * which accompanies this distribution, and is available at
 * https://www.eclipse.org/legal/epl-2.0/
 *
 *******************************************************************************/

package org.eclipse.tcf.internal.dap;

import java.util.LinkedList;
import java.util.concurrent.CompletableFuture;

import org.eclipse.tcf.dap.IRequest;
import org.eclipse.tcf.dap.IStatus;
import org.eclipse.tcf.dap.ITCFObject;
import org.eclipse.tcf.protocol.IChannel;
import org.eclipse.tcf.protocol.Protocol;
import org.eclipse.tcf.util.TCFDataCache;


/**
 * TCFNode is base class for all TCF debug model elements.
 */
public abstract class TCFNode implements ITCFObject, Comparable<TCFNode> {


    protected final String id;
    protected final TCFNode parent;
    protected final TCFModel model;
    protected final TCFLaunch launch;
    protected final IChannel channel;

    private boolean disposed;

    /**
     * An extension of TCFDataCache class that is automatically disposed when the node is disposed.
     */
    protected abstract class TCFData<V> extends TCFDataCache<V> {

        TCFData(IChannel channel) {
            super(channel);
            addDataCache(this);
        }

        @Override
        public void dispose() {
            removeDataCache(this);
            super.dispose();
        }
    }

    private LinkedList<TCFDataCache<?>> caches;

    /**
     * Constructor for a root node. There should be exactly one root in the model.
     * @param model
     */
    protected TCFNode(TCFModel model) {
        id = null;
        parent = null;
        launch = model.getLaunch();
        channel = model.getChannel();
        this.model = model;
    }

    /**
     * Constructor for a node other then root. Node ID must be unique.
     * @param parent - parent node.
     * @param id - node ID.
     */
    protected TCFNode(TCFNode parent, String id) {
        assert Protocol.isDispatchThread();
        assert parent != null;
        assert id != null;
        assert !parent.disposed;
        this.parent = parent;
        this.id = id;
        model = parent.model;
        model.addNode(id, this);
        launch = model.getLaunch();
        channel = model.getChannel();
    }

    /**
     * Register a data cache object that caches data for this node.
     * @param c - a TCFData object.
     */
    final void addDataCache(TCFDataCache<?> c) {
        if (caches == null) caches = new LinkedList<TCFDataCache<?>>();
        caches.add(c);
    }

    /**
     * Unregister a data cache object that caches children for this node.
     * @param c - a TCFData object.
     */
    final void removeDataCache(TCFDataCache<?> c) {
        if (caches != null) caches.remove(c);
    }

    /**
     * Flush (reset) all node data caches.
     */
    void flushAllCaches() {
        if (caches == null) return;
        for (TCFDataCache<?> c : caches) c.reset();
    }

    /**
     * Dispose this node and its children. The node is removed from the model.
     */
    void dispose() {
        assert !disposed;
        while (caches != null && caches.size() > 0) {
            caches.getLast().dispose();
        }
        if (parent != null && parent.caches != null) {
            for (TCFDataCache<?> c : parent.caches) {
                if (c instanceof TCFChildren) ((TCFChildren)c).onNodeDisposed(id);
            }
        }
        if (id != null) {
            assert model.getNode(id) == this;
            model.removeNode(id);
        }
        disposed = true;
    }

    /**
     * Check if node is disposed.
     * @return true if disposed.
     */
    public final boolean isDisposed() {
        return disposed;
    }

    /**
     * Get TCFModel that owns this node.
     * @return TCFModel object
     */
    public TCFModel getModel() {
        return model;
    }

    /**
     * Get IChannel of TCFModel that owns this node.
     * @return IChannel object
     */
    public IChannel getChannel() {
        return channel;
    }

    /**
     * Get TCF ID of the node.
     * @return TCF ID
     */
    public String getID() {
        return id;
    }

    /**
     * Get parent node.
     * @return parent node or null if the node is a root
     */
    public final TCFNode getParent() {
        return parent;
    }


    /**
     * Get parent node in given presentation context.
    // * @param ctx - presentation context.
     * @return parent node or null if the node is a root
     */
    //TODO:
//    public TCFNode getParent(IPresentationContext ctx) {
//        assert Protocol.isDispatchThread();
//        return parent;
//    }

    protected boolean getData(final CompletableFuture<Boolean> update, Runnable done) {
        // update.setChildCount(0);
        update.complete(true);
        return true;
    }
    /**
     * Retrieve node label for a presentation context.
     * // @param update - label update request.
     */
    String getData( Runnable runnable){
        return "...";
    }
    final String getLabel( Runnable runnable) {
        if (!disposed && channel.getState() == IChannel.STATE_OPEN) {
            return getData(runnable);
        } else {
            return "...";
        }
    }



    /**
     * Retrieve children count for a presentation context.
     * @param update - children count update request.
     */
    //TODO:
//    final void update(final IChildrenCountUpdate update) {
//        new TCFRunnable(model, update) {
//            public void run() {
//                if (!done) {
//                    if (!update.isCanceled()) {
//                        if (!disposed && channel.getState() == IChannel.STATE_OPEN) {
//                            if (!getData(update, this)) return;
//                        }
//                        else {
//                            update.setChildCount(0);
//                        }
//                        update.setStatus(Status.OK_STATUS);
//                    }
//                    done();
//                }
//            }
//        };
//    }

    /**
     * Retrieve children for a presentation context.
     * @param update - children update request.
     */
    //TODO:
//    final void update(final IChildrenUpdate update) {
//        new TCFRunnable(model, update) {
//            public void run() {
//                if (!done) {
//                    if (!update.isCanceled()) {
//                        if (!disposed && channel.getState() == IChannel.STATE_OPEN) {
//                            if (!getLockedData(update, this)) return;
//                        }
//                        update.setStatus(Status.OK_STATUS);
//                    }
//                    done();
//                }
//            }
//        };
//    }

    /**
     * Check if the node has children in a presentation context.
     * @param update - "has children" update request.
     */
    //TODO:
//    final void update(final IHasChildrenUpdate update) {
//        new TCFRunnable(model, update) {
//            public void run() {
//                if (!done) {
//                    if (!update.isCanceled()) {
//                        if (!disposed && channel.getState() == IChannel.STATE_OPEN) {
//                            if (!getLockedData(update, this)) return;
//                        }
//                        else {
//                            update.setHasChilren(false);
//                        }
//                        update.setStatus(Status.OK_STATUS);
//                    }
//                    done();
//                }
//            }
//        };
//    }

    /**
     * Retrieve node label for a presentation context.
     * @param update - label update request.
     */
    //TODO:
//    final void update(final ILabelUpdate update) {
//        new TCFRunnable(model, update) {
//            public void run() {
//                if (!done) {
//                    if (!update.isCanceled()) {
//                        if (!disposed && channel.getState() == IChannel.STATE_OPEN) {
//                            if (!getLockedData(update, this)) return;
//                            getFontData(update, update.getPresentationContext().getId());
//                        }
//                        else {
//                            update.setLabel("...", 0);
//                        }
//                        update.setStatus(Status.OK_STATUS);
//                    }
//                    done();
//                }
//            }
//        };
//    }

    /**
     * Retrieve viewer input object for a presentation context.
     * Allows a view to translate the active debug context into an appropriate viewer input element.
     * @param update - input update request.
     */
    //TODO:
//    final void update(final IViewerInputUpdate update) {
//        new TCFRunnable(model, update) {
//            public void run() {
//                if (!done) {
//                    if (!update.isCanceled()) {
//                        if (!disposed && channel.getState() == IChannel.STATE_OPEN) {
//                            if (!getData(update, this)) return;
//                        }
//                        else {
//                            update.setInputElement(TCFNode.this);
//                        }
//                        update.setStatus(Status.OK_STATUS);
//                    }
//                    done();
//                }
//            }
//        };
//    }

    /**
     * Retrieve children count for a presentation context.
     * If the context is locked, return snapshot data.
     * Otherwise return live data from the target.
     * The method is always called on TCF dispatch thread.
     * @param update - children count update request.
     * @param done - client call back interface, during data waiting it is
     * called every time new portion of data becomes available.
     * @return false if waiting data retrieval, true if all done.
     */
    //TODO:
//    final boolean getLockedData(IChildrenCountUpdate update, Runnable done) {
//        TCFSnapshot snapshot = model.getSnapshot(update.getPresentationContext());
//        if (snapshot != null) return snapshot.getData(update, this, done);
//        return getData(update, done);
//    }

    /**
     * Retrieve children for a presentation context.
     * If the context is locked, return snapshot data.
     * Otherwise return live data from the target.
     * The method is always called on TCF dispatch thread.
     * @param update - children update request.
     * @param done - client call back interface, during data waiting it is
     * called every time new portion of data becomes available.
     * @return false if waiting data retrieval, true if all done.
     */
    //TODO:
//    final boolean getLockedData(IChildrenUpdate update, Runnable done) {
//        TCFSnapshot snapshot = model.getSnapshot(update.getPresentationContext());
//        if (snapshot != null) return snapshot.getData(update, this, done);
//        return getData(update, done);
//    }

    /**
     * Check if the node has children in a presentation context.
     * If the context is locked, return snapshot data.
     * Otherwise return live data from the target.
     * The method is always called on TCF dispatch thread.
     * @param update - "has children" update request.
     * @param done - client call back interface, during data waiting it is
     * called every time new portion of data becomes available.
     * @return false if waiting data retrieval, true if all done.
     */
    //TODO:
//    final boolean getLockedData(IHasChildrenUpdate update, Runnable done) {
//        TCFSnapshot snapshot = model.getSnapshot(update.getPresentationContext());
//        if (snapshot != null) return snapshot.getData(update, this, done);
//        return getData(update, done);
//    }

    /**
     * Retrieve node label for a presentation context.
     * If the context is locked, return snapshot data.
     * Otherwise return live data from the target.
     * The method is always called on TCF dispatch thread.
     * @param update - label update request.
     * @param done - client call back interface, during data waiting it is
     * called every time new portion of data becomes available.
     * @return false if waiting data retrieval, true if all done.
     */
    //TODO:
//    final boolean getLockedData(ILabelUpdate update, Runnable done) {
//        TCFSnapshot snapshot = model.getSnapshot(update.getPresentationContext());
//        if (snapshot != null) return snapshot.getData(update, this, done);
//        return getData(update, done);
//    }

    /**
     * Retrieve children count for a presentation context.
     * The method is always called on TCF dispatch thread.
     * @param update - children count update request.
     * @param done - client call back interface, during data waiting it is
     * called every time new portion of data becomes available.
     * @return false if waiting data retrieval, true if all done.
     */
    //TODO:
//    protected boolean getData(IChildrenCountUpdate update, Runnable done) {
//        update.setChildCount(0);
//        return true;
//    }

    /**
     * Retrieve children for a presentation context.
     * The method is always called on TCF dispatch thread.
     * @param update - children update request.
     * @param done - client call back interface, during data waiting it is
     * called every time new portion of data becomes available.
     * @return false if waiting data retrieval, true if all done.
     */
    //TODO:
//    protected boolean getData(IChildrenUpdate update, Runnable done) {
//        return true;
//    }

    /**
     * Check if the node has children in a presentation context.
     * The method is always called on TCF dispatch thread.
     * @param update - "has children" update request.
     * @param done - client call back interface, during data waiting it is
     * called every time new portion of data becomes available.
     * @return false if waiting data retrieval, true if all done.
     */
    //TODO:
//    protected boolean getData(IHasChildrenUpdate update, Runnable done) {
//        update.setHasChilren(false);
//        return true;
//    }

    /**
     * Retrieve node label for a presentation context.
     * The method is always called on TCF dispatch thread.
     * @param update - label update request.
     * @param done - client call back interface, during data waiting it is
     * called every time new portion of data becomes available.
     * @return false if waiting data retrieval, true if all done.
     */
    //TODO:
//    protected boolean getData(ILabelUpdate update, Runnable done) {
//        update.setLabel(id, 0);
//        return true;
//    }

    /**
     * Retrieve viewer input object for a presentation context.
     * Allows a view to translate the active debug context into an appropriate viewer input element.
     * The method is always called on TCF dispatch thread.
     * @param update - view input update request.
     * @param done - client call back interface, during data waiting it is
     * called every time new portion of data becomes available.
     * @return false if waiting data retrieval, true if all done.
     */
    //TODO:
//    protected boolean getData(IViewerInputUpdate update, Runnable done) {
//        update.setInputElement(this);
//        return true;
//    }

    /**
     * Get FontData info for node label update.
     * @param update - label update request.
     */
    //TODO:
//    protected void getFontData(ILabelUpdate update, String view_id) {
//        FontData fd = TCFModelFonts.getNormalFontData(view_id);
//        String[] cols = update.getColumnIds();
//        if (cols == null || cols.length == 0) {
//            update.setFontData(fd, 0);
//        }
//        else {
//            for (int i = 0; i < cols.length; i++) update.setFontData(fd, i);
//        }
//    }

    /**
     * Creates and stores a IMemento for the node.
     * A request should be cancelled if a IMemento is not supported.
     * @param request Specifies IMemento.
     */
    //TODO:
//    public void encodeElement(IElementMementoRequest request) {
//        request.getMemento().putString("TCF.ID", id);
//    }

    /**
     * Determines whether a IMemento represents this node.
     * @param request Specifies previously created IMemento.
     */
    //TODO:
//    public void compareElements(IElementCompareRequest request) {
//        request.setEqual(id.equals(request.getMemento().getString("TCF.ID")));
//    }

    /*--------------------------------------------------------------------------------------*/
    /* Misc                                                                                 */

    /**
     * Flush all caches and repaint the node and its children in presentation context of given part.
     * @param part - workbench part that needs to be refreshed.
     */
    //TODO:
//    public void refresh(IWorkbenchPart part) {
//        model.flushAllCaches();
//        for (TCFModelProxy p : model.getModelProxies()) {
//            if (p.getPresentationContext().getPart() != part) continue;
//            p.addDelta(this, IModelDelta.STATE | IModelDelta.CONTENT);
//        }
//    }

    /**
     * Compare two nodes.
     * Extensions of TCFNode are expected to override this method.
     */
    public int compareTo(TCFNode n) {
        return id.compareTo(n.id);
    }

    /**
     * Returns a simple human readable string representation of the node.
     */
    public String toString() {
        String s = "[" + Integer.toHexString(hashCode()) + "] " + id;
        if (disposed) s += ", disposed";
        return s;
    }
}
